const {is_node} = require("./env");

const {
    FormatError
} = require("./err");


const {
    CFFEncodingMap,
    CFFDictDataMap,
    CFFDictPrivateDataMap
} = require("./cff");


function readCharstringEncoding(aString) {
  if (!aString) {
    return '';
  }
  var charstringTokens = [];
  var count = aString.length;
  for (var i = 0; i < count; ) {
    var value = aString[i++] | 0;
    var token = null;
    if (value < 0) {
      continue;
    } else if (value <= 11) {
      token = CFFEncodingMap[value];
    } else if (value === 12) {
      token = CFFEncodingMap[value][aString[i++]];
    } else if (value <= 18) {
      token = CFFEncodingMap[value];
    } else if (value <= 20) {
      ++i; // var mask = aString[i++];
      token = CFFEncodingMap[value];
    } else if (value <= 27) {
      token = CFFEncodingMap[value];
    } else if (value === 28) {
      token = aString[i++] << 8 | aString[i++];
    } else if (value <= 31) {
      token = CFFEncodingMap[value];
    } else if (value < 247) {
      token = parseInt(value, 10) - 139;
    } else if (value < 251) {
      token = (value - 247) * 256 + aString[i++] + 108;
    } else if (value < 255) {
      token = -(value - 251) * 256 - aString[i++] - 108;
    } else { // value === 255
      token = aString[i++] << 24 | aString[i++] << 16 |
              aString[i++] << 8 | aString[i];
    }
    charstringTokens.push(token);
  }
  return charstringTokens;
}

function readCharset(aStream, aCharstrings,CFFStrings) {
  ////
  CFFStrings = CFFStrings??globalThis.CFFStrings;
  ////  
  var charset = {};
  var format = aStream.getByte();
  var count = aCharstrings.length - 1;
  var i, sid;
  if (format === 0) {
    charset['.notdef'] = readCharstringEncoding(aCharstrings[0]);

    for (i = 1; i < count + 1; i++) {
      sid = aStream.getByte() << 8 | aStream.getByte();
      charset[CFFStrings[sid]] = readCharstringEncoding(aCharstrings[i]);
    }
  } else if (format === 1) {
    for (i = 1; i < count + 1; i++) {
      var first = aStream.getByte();
      first = (first << 8) | aStream.getByte();
      var numLeft = aStream.getByte();
      for (var j = 0; j <= numLeft; j++) {
        sid = first++;
        charset[CFFStrings[sid]] = readCharstringEncoding(aCharstrings[j]);
      }
    }
  } else {
    throw new FormatError('Invalid charset format');
  }
  return charset;
}


function readFontDictData(aString, aMap) {
  var fontDictDataTokens = [];
  var count = aString.length;
  for (var i = 0; i < count; ) {
    var value = aString[i++] | 0;
    var token = null;
    if (value === 12) {
      token = aMap[value][aString[i++]];
    } else if (value === 28) {
      token = aString[i++] << 8 | aString[i++];
    } else if (value === 29) {
      token = aString[i++] << 24 |
              aString[i++] << 16 |
              aString[i++] << 8 |
              aString[i++];
    } else if (value === 30) {
      token = '';
      var parsed = false;
      while (!parsed) {
        var octet = aString[i++];
        var nibbles = [parseInt(octet / 16, 10), parseInt(octet % 16, 10)];
        for (var j = 0; j < nibbles.length; j++) {
          var nibble = nibbles[j];
          switch (nibble) {
            case 0xA:
              token += '.';
              break;
            case 0xB:
              token += 'E';
              break;
            case 0xC:
              token += 'E-';
              break;
            case 0xD:
              break;
            case 0xE:
              token += '-';
              break;
            case 0xF:
              parsed = true;
              break;
            default:
              token += nibble;
              break;
          }
        }
      }
      token = parseFloat(token);
    } else if (value <= 31) {
      token = aMap[value];
    } else if (value <= 246) {
      token = parseInt(value, 10) - 139;
    } else if (value <= 250) {
      token = (value - 247) * 256 + aString[i++] + 108;
    } else if (value <= 254) {
      token = -(value - 251) * 256 - aString[i++] - 108;
    } else if (value === 255) {
      throw new FormatError('255 is not a valid DICT command');
    }
    fontDictDataTokens.push(token);
  }
  return fontDictDataTokens;
}


function readFontIndexData(aStream, aIsByte) {
  var count = aStream.getByte() << 8 | aStream.getByte();
  var offsize = aStream.getByte();
  function getNextOffset() {
    switch (offsize) {
      case 0:
        return 0;
      case 1:
        return aStream.getByte();
      case 2:
        return aStream.getByte() << 8 | aStream.getByte();
      case 3:
        return aStream.getByte() << 16 | aStream.getByte() << 8 |
               aStream.getByte();
      case 4:
      return aStream.getByte() << 24 | aStream.getByte() << 16 |
             aStream.getByte() << 8 | aStream.getByte();
    }
    throw new FormatError(offsize + ' is not a valid offset size');
  }
  var offsets = [];
  var i;
  for (i = 0; i < count + 1; i++) {
    offsets.push(getNextOffset());
  }
  //dump('Found ' + count + ' objects at offsets :' +offsets + ' (offsize: ' + offsize + ')');
  // Now extract the objects
  var relativeOffset = aStream.pos;
  var objects = [];
  for (i = 0; i < count; i++) {
    var offset = offsets[i];
    aStream.pos = relativeOffset + offset - 1;

    var data = [];
    var length = offsets[i + 1] - 1;
    for (var j = offset - 1; j < length; j++) {
      data.push(aIsByte ? aStream.getByte() : aStream.getChar());
    }
    objects.push(data);
  }
  return objects;
}

////

const {
    isNum
} = require("./is");


function parseAsToken(font,aString, aMap,CFFStrings) {
   ////
   CFFStrings = CFFStrings??globalThis.CFFStrings;
   ////
    var decoded = readFontDictData(aString, aMap);
    var stack = [];
    var count = decoded.length;
    for (var i = 0; i < count; i++) {
      var token = decoded[i];
      if (isNum(token)) {
        stack.push(token);
      } else {
        switch (token.operand) {
          case 'SID':
            font.set(token.name, CFFStrings[stack.pop()]);
            break;
          case 'number number':
            font.set(token.name, {
              offset: stack.pop(),
              size: stack.pop()
            });
            break;
          case 'boolean':
            font.set(token.name, stack.pop());
            break;
          case 'delta':
            font.set(token.name, stack.pop());
            break;
          default:
            if (token.operand && token.operand.length) {
              var array = [];
              for (var j = 0; j < token.operand.length; j++) {
                array.push(stack.pop());
              }
              font.set(token.name, array);
            } else {
              font.set(token.name, stack.pop());
            }
            break;
        }
      }
    }
}


function type2ParserParse(font,aStream,CFFStrings) {
      ////
      CFFStrings = CFFStrings??globalThis.CFFStrings;
      ////
      font.set('major', aStream.getByte());
      font.set('minor', aStream.getByte());
      font.set('hdrSize', aStream.getByte());
      font.set('offsize', aStream.getByte());
      font.set('Names', readFontIndexData(aStream));
      var topDict = readFontIndexData(aStream, true);
      var strings = readFontIndexData(aStream);
      // Fill up the Strings dictionary with the new unique strings
      var i;
      for (i = 0; i < strings.length; i++) {
          CFFStrings.push(strings[i].join(''));
      }
      // Parse the TopDict operator
      var count = topDict.length;
      for (i = 0; i < count; i++) {
          parseAsToken(font,topDict[i], CFFDictDataMap,CFFStrings);
      }
      ////
      var subrs = readFontIndexData(aStream, true);
      // Reading Private Dict
      var priv = font.get('Private');
      aStream.pos = priv.offset;
      var privateDict = [];
      for (i = 0; i < priv.size; i++) {
          privateDict.push(aStream.getByte());
      }
      parseAsToken(font,privateDict, CFFDictPrivateDataMap,CFFStrings);
      // Read CharStrings Index
      var charStringsOffset = font.get('CharStrings');
      aStream.pos = charStringsOffset;
      var charStrings = readFontIndexData(aStream, true);
      // Read Charset
      var charsetEntry = font.get('charset');
      if (charsetEntry === 0) {
          throw new FormatError('Need to support CFFISOAdobeCharset');
      } else if (charsetEntry === 1) {
          throw new FormatError('Need to support CFFExpert');
      } else if (charsetEntry === 2) {
          throw new FormatError('Need to support CFFExpertSubsetCharset');
      } else {
          aStream.pos = charsetEntry;
          readCharset(aStream, charStrings,CFFStrings);
      }
}



////
const {
    Dict
} = require("./primitives");

const {
    Stream
} = require("./stream");


function load_font_file_to_stream(aFilePath) {
    if(is_node()) {
        let buf = fs.readFileSync(aFilePath);
        return(new Stream(buf));
    } else {
        let xhr = new XMLHttpRequest();
        xhr.open('GET', aFilePath, false);
        xhr.responseType = 'arraybuffer';
        xhr.expected = document.URL.startsWith('file:') ? 0 : 200;
        xhr.send(null);
        return(new Stream(xhr.response));
    }
}


class Type2Parser {
    #font = new Dict(null)
    #CFFStrings = globalThis.CFFStrings
    constructor(aFilePath,CFFStrings) {
        this.#CFFStrings = CFFStrings??globalThis.CFFStrings
        this.data = load_font_file_to_stream(aFilePath);
    }
    parse(aStream) {
       aStream = aStream??this.data;
       return(type2ParserParse(this.#font,aStream,this.#CFFStrings)) 
    }
}


////


function writeToFile(aBytes, aFilePath) {
    if(is_node()) {
        const fs = require("fs");
        fs.writeFileSync(aFilePath,aBytes)
    } else {
        let blob = new Blob(aBytes);
        let ourl = URL.createObjectURL(blob);
        return(ourl)
    }
}

////
// See https://developer.mozilla.org/en-US/docs/Web/CSS/font-style.
const DEFAULT_CSS_FONT_OBLIQUE = "14";
// See https://developer.mozilla.org/en-US/docs/Web/CSS/font-weight.
const DEFAULT_CSS_FONT_WEIGHT = "400";
const CSS_FONT_WEIGHT_VALUES = new Set([
    "100",
    "200",
    "300",
    "400",
    "500",
    "600",
    "700",
    "800",
    "900",
    "1000",
    "normal",
    "bold",
    "bolder",
    "lighter",
]);


function validateCSSFont(cssFontInfo) {
  const { fontFamily, fontWeight, italicAngle } = cssFontInfo;
  // See https://developer.mozilla.org/en-US/docs/Web/CSS/string.
  if (/^".*"$/.test(fontFamily)) {
    if (/[^\\]"/.test(fontFamily.slice(1, fontFamily.length - 1))) {
      return false;
    }
  } else if (/^'.*'$/.test(fontFamily)) {
    if (/[^\\]'/.test(fontFamily.slice(1, fontFamily.length - 1))) {
      return false;
    }
  } else {
    // See https://developer.mozilla.org/en-US/docs/Web/CSS/custom-ident.
    for (const ident of fontFamily.split(/[ \t]+/)) {
      if (/^(\d|(-(\d|-)))/.test(ident) || !/^[\w-\\]+$/.test(ident)) {
        return false;
      }
    }
  }
  const weight = fontWeight ? fontWeight.toString() : "";
  cssFontInfo.fontWeight = CSS_FONT_WEIGHT_VALUES.has(weight)
    ? weight
    : DEFAULT_CSS_FONT_WEIGHT;
  
  const angle = parseFloat(italicAngle);
  cssFontInfo.italicAngle =
    isNaN(angle) || angle < -90 || angle > 90
      ? DEFAULT_CSS_FONT_OBLIQUE
      : italicAngle.toString();
  return true;
}


module.exports = {
    readCharstringEncoding,
    readCharset,
    readFontDictData,
    readFontIndexData,
    ////
    parseAsToken,
    type2ParserParse,
    load_font_file_to_stream,
    Type2Parser,
    ////
    writeToFile,
    ////
    DEFAULT_CSS_FONT_OBLIQUE,
    DEFAULT_CSS_FONT_WEIGHT,
    CSS_FONT_WEIGHT_VALUES,
    validateCSSFont,
    ////

}
